@@ -33,6 +33,7 @@ struct _IBusM17NEngineClass {
     guint preedit_foreground;
     guint preedit_background;
     gint preedit_underline;
+    IBusPreeditFocusMode preedit_focus_mode;
     gint lookup_table_orientation;
 
     MInputMethod *im;
@@ -43,11 +44,7 @@ static void ibus_m17n_engine_class_init     (IBusM17NEngineClass    *klass);
 static void ibus_m17n_config_value_changed  (IBusConfig             *config,
                                              const gchar            *section,
                                              const gchar            *name,
-#if IBUS_CHECK_VERSION(1,3,99)
                                              GVariant               *value,
-#else
-                                             GValue                 *value,
-#endif  /* !IBUS_CHECK_VERSION(1,3,99) */
                                              IBusM17NEngineClass    *klass);
 
 static GObject*
@@ -215,7 +212,7 @@ ibus_m17n_engine_class_init (IBusM17NEngineClass *klass)
     IBusEngineClass *engine_class = IBUS_ENGINE_CLASS (klass);
     gchar *engine_name, *lang = NULL, *name = NULL;
     IBusM17NEngineConfig *engine_config;
-    gchar *hex;
+    GVariant *values;
 
     if (parent_class == NULL)
         parent_class = (IBusEngineClass *) g_type_class_peek_parent (klass);
@@ -251,44 +248,64 @@ ibus_m17n_engine_class_init (IBusM17NEngineClass *klass)
     g_free (lang);
     g_free (name);
 
-    /* configurations are per class */
-    klass->preedit_foreground = INVALID_COLOR;
-    klass->preedit_background = INVALID_COLOR;
-    klass->preedit_underline = IBUS_ATTR_UNDERLINE_NONE;
-    klass->lookup_table_orientation = IBUS_ORIENTATION_SYSTEM;
-
     engine_config = ibus_m17n_get_engine_config (engine_name);
     g_free (engine_name);
 
-    if (ibus_m17n_config_get_string (config,
-                                     klass->config_section,
-                                     "preedit_foreground",
-                                     &hex)) {
-        klass->preedit_foreground = ibus_m17n_parse_color (hex);
-        g_free (hex);
-    } else if (engine_config->preedit_highlight)
-        klass->preedit_foreground = PREEDIT_FOREGROUND;
+    /* configurations are per class */
+    klass->preedit_foreground = engine_config->preedit_highlight ?
+        PREEDIT_FOREGROUND :
+        INVALID_COLOR;
+    klass->preedit_background = engine_config->preedit_highlight ?
+        PREEDIT_BACKGROUND :
+        INVALID_COLOR;
+    klass->preedit_underline = IBUS_ATTR_UNDERLINE_NONE;
+    klass->preedit_focus_mode = engine_config->preedit_highlight ?
+        IBUS_ENGINE_PREEDIT_COMMIT :
+        IBUS_ENGINE_PREEDIT_CLEAR;
+    klass->lookup_table_orientation = IBUS_ORIENTATION_SYSTEM;
 
-    if (ibus_m17n_config_get_string (config,
-                                     klass->config_section,
-                                     "preedit_background",
-                                     &hex)) {
-        klass->preedit_background = ibus_m17n_parse_color (hex);
-        g_free (hex);
-    } else if (engine_config->preedit_highlight)
-        klass->preedit_background = PREEDIT_BACKGROUND;
+    ibus_m17n_engine_config_free (engine_config);
 
-    if (!ibus_m17n_config_get_int (config,
-                                   klass->config_section,
-                                   "preedit_underline",
-                                   &klass->preedit_underline))
-        klass->preedit_underline = IBUS_ATTR_UNDERLINE_NONE;
+    values = ibus_config_get_values (config,
+                                     klass->config_section);
+    if (values != NULL) {
+        GVariant *value;
 
-    if (!ibus_m17n_config_get_int (config,
-                                   klass->config_section,
-                                   "lookup_table_orientation",
-                                   &klass->lookup_table_orientation))
-        klass->lookup_table_orientation = IBUS_ORIENTATION_SYSTEM;
+        value = g_variant_lookup_value (values,
+                                        "preedit_foreground",
+                                        G_VARIANT_TYPE_STRING);
+        if (value != NULL) {
+            const gchar *hex = g_variant_get_string (value, NULL);
+            klass->preedit_foreground = ibus_m17n_parse_color (hex);
+            g_variant_unref (value);
+        }
+
+        value = g_variant_lookup_value (values,
+                                        "preedit_background",
+                                        G_VARIANT_TYPE_STRING);
+        if (value != NULL) {
+            const gchar *hex = g_variant_get_string (value, NULL);
+            klass->preedit_background = ibus_m17n_parse_color (hex);
+            g_variant_unref (value);
+        }
+
+        value = g_variant_lookup_value (values,
+                                        "preedit_underline",
+                                        G_VARIANT_TYPE_INT32);
+        if (value != NULL) {
+            klass->preedit_background = g_variant_get_int32 (value);
+            g_variant_unref (value);
+        }
+
+        value = g_variant_lookup_value (values,
+                                        "lookup_table_orientation",
+                                        G_VARIANT_TYPE_INT32);
+        if (value != NULL) {
+            klass->lookup_table_orientation = g_variant_get_int32 (value);
+            g_variant_unref (value);
+        }
+        g_variant_unref (values);
+    }
 
     g_signal_connect (config, "value-changed",
                       G_CALLBACK(ibus_m17n_config_value_changed),
@@ -297,44 +314,32 @@ ibus_m17n_engine_class_init (IBusM17NEngineClass *klass)
     klass->im = NULL;
 }
 
-#if IBUS_CHECK_VERSION(1,3,99)
-#define _g_variant_get_string g_variant_get_string
-#define _g_variant_get_int32 g_variant_get_int32
-#else
-#define _g_variant_get_string(value, length) g_value_get_string(value)
-#define _g_variant_get_int32 g_value_get_int
-#endif  /* !IBUS_CHECK_VERSION(1,3,99) */
-
 static void
 ibus_m17n_config_value_changed (IBusConfig          *config,
                                 const gchar         *section,
                                 const gchar         *name,
-#if IBUS_CHECK_VERSION(1,3,99)
                                 GVariant            *value,
-#else
-                                GValue              *value,
-#endif  /* !IBUS_CHECK_VERSION(1,3,99) */
                                 IBusM17NEngineClass *klass)
 {
     if (g_strcmp0 (section, klass->config_section) == 0) {
         if (g_strcmp0 (name, "preedit_foreground") == 0) {
-            const gchar *hex = _g_variant_get_string (value, NULL);
+            const gchar *hex = g_variant_get_string (value, NULL);
             guint color;
             color = ibus_m17n_parse_color (hex);
             if (color != INVALID_COLOR) {
                 klass->preedit_foreground = color;
             }
         } else if (g_strcmp0 (name, "preedit_background") == 0) {
-            const gchar *hex = _g_variant_get_string (value, NULL);
+            const gchar *hex = g_variant_get_string (value, NULL);
             guint color;
             color = ibus_m17n_parse_color (hex);
             if (color != INVALID_COLOR) {
                 klass->preedit_background = color;
             }
         } else if (g_strcmp0 (name, "preedit_underline") == 0) {
-            klass->preedit_underline = _g_variant_get_int32 (value);
+            klass->preedit_underline = g_variant_get_int32 (value);
         } else if (g_strcmp0 (name, "lookup_table_orientation") == 0) {
-            klass->lookup_table_orientation = _g_variant_get_int32 (value);
+            klass->lookup_table_orientation = g_variant_get_int32 (value);
         }
     }
 }
@@ -492,10 +497,11 @@ ibus_m17n_engine_update_preedit (IBusM17NEngine *m17n)
                                         klass->preedit_background, 0, -1);
         ibus_text_append_attribute (text, IBUS_ATTR_TYPE_UNDERLINE,
                                     klass->preedit_underline, 0, -1);
-        ibus_engine_update_preedit_text ((IBusEngine *) m17n,
-                                         text,
-                                         m17n->context->cursor_pos,
-                                         mtext_len (m17n->context->preedit) > 0);
+        ibus_engine_update_preedit_text_with_mode ((IBusEngine *) m17n,
+                                                   text,
+                                                   m17n->context->cursor_pos,
+                                                   mtext_len (m17n->context->preedit) > 0,
+                                                   klass->preedit_focus_mode);
     }
 }
 
@@ -524,7 +530,7 @@ ibus_m17n_engine_commit_string (IBusM17NEngine *m17n,
                 
    Since IBus engines are supposed to be cross-platform, the code
    should go into IBus core, instead of ibus-m17n. */
-MSymbol
+static MSymbol
 ibus_m17n_key_event_to_symbol (guint keycode,
                                guint keyval,
                                guint modifiers)
@@ -566,12 +572,13 @@ ibus_m17n_key_event_to_symbol (guint keycode,
         g_string_append_c (keysym, c);
     }
     else {
-        mask |= modifiers & (IBUS_CONTROL_MASK | IBUS_SHIFT_MASK);
-        g_string_append (keysym, ibus_keyval_name (keyval));
-        if (keysym->len == 0) {
+        const gchar *name = ibus_keyval_name (keyval);
+        if (name == NULL) {
             g_string_free (keysym, TRUE);
             return Mnil;
         }
+        mask |= modifiers & (IBUS_CONTROL_MASK | IBUS_SHIFT_MASK);
+        g_string_append (keysym, name);
     }
 
     mask |= modifiers & (IBUS_MOD1_MASK |
@@ -676,7 +683,10 @@ ibus_m17n_engine_focus_out (IBusEngine *engine)
 {
     IBusM17NEngine *m17n = (IBusM17NEngine *) engine;
 
-    ibus_m17n_engine_process_key (m17n, Minput_focus_out);
+    /* To make ibus_engine_update_preedit_text_with_mode work
+       properly, we just reset the IC instead of passing Mfocus_out to
+       m17n-lib. */
+    minput_reset_ic (m17n->context);
 
     parent_class->focus_out (engine);
 }
@@ -698,11 +708,9 @@ ibus_m17n_engine_enable (IBusEngine *engine)
 
     parent_class->enable (engine);
 
-#ifdef HAVE_IBUS_ENGINE_GET_SURROUNDING_TEXT
     /* Issue a dummy ibus_engine_get_surrounding_text() call to tell
        input context that we will use surrounding-text. */
     ibus_engine_get_surrounding_text (engine, NULL, NULL, NULL);
-#endif  /* HAVE_IBUS_ENGINE_GET_SURROUNDING_TEXT */
 }
 
 static void
@@ -814,8 +822,15 @@ ibus_m17n_engine_update_lookup_table (IBusM17NEngine *m17n)
             ibus_lookup_table_set_page_size (m17n->table, mtext_len (mt));
 
             buf = ibus_m17n_mtext_to_ucs4 (mt, &nchars);
-            for (i = 0; i < nchars; i++) {
-                ibus_lookup_table_append_candidate (m17n->table, ibus_text_new_from_unichar (buf[i]));
+            g_warn_if_fail (buf != NULL);
+
+            for (i = 0; buf != NULL && i < nchars; i++) {
+                IBusText *text = ibus_text_new_from_unichar (buf[i]);
+                if (text == NULL) {
+                    text = ibus_text_new_from_printf ("INVCODE=U+%04"G_GINT32_FORMAT"X", buf[i]);
+                    g_warn_if_reached ();
+                }
+                ibus_lookup_table_append_candidate (m17n->table, text);
             }
             g_free (buf);
         }
@@ -832,9 +847,15 @@ ibus_m17n_engine_update_lookup_table (IBusM17NEngine *m17n)
                 mtext = (MText *) mplist_value (p);
                 buf = ibus_m17n_mtext_to_utf8 (mtext);
                 if (buf) {
-                    ibus_lookup_table_append_candidate (m17n->table, ibus_text_new_from_string (buf));
+                    ibus_lookup_table_append_candidate (m17n->table,
+                        ibus_text_new_from_string (buf));
                     g_free (buf);
                 }
+                else {
+                    ibus_lookup_table_append_candidate (m17n->table,
+                        ibus_text_new_from_static_string ("NULL"));
+                    g_warn_if_reached();
+                }
             }
         }
 
@@ -916,9 +937,6 @@ ibus_m17n_engine_callback (MInputContext *context,
     }
     else if (command == Minput_reset) {
     }
-    /* ibus_engine_get_surrounding_text is only available in the current
-       git master (1.3.99+) */
-#ifdef HAVE_IBUS_ENGINE_GET_SURROUNDING_TEXT
     else if (command == Minput_get_surrounding_text &&
              (((IBusEngine *) m17n)->client_capabilities &
               IBUS_CAP_SURROUNDING_TEXT) != 0) {
@@ -956,7 +974,6 @@ ibus_m17n_engine_callback (MInputContext *context,
         mplist_set (m17n->context->plist, Mtext, surround);
         m17n_object_unref (surround);
     }
-#endif  /* !HAVE_IBUS_ENGINE_GET_SURROUNDING_TEXT */
     else if (command == Minput_delete_surrounding_text &&
              (((IBusEngine *) m17n)->client_capabilities &
               IBUS_CAP_SURROUNDING_TEXT) != 0) {
